/*****************************************************************************/
/*
                                SWISHesi.c

              +---------------------------------------------+
              | THIS IS AN IN-DEVELOPMENT COPY OF VERSION 1 |
              +---------------------------------------------+

SWISH-E Search Interface.
Can be pronounced swish-easy :^)

This is a simple CGI interface to SWISH-E indices, allowing searches of indexed
documents to be initiated via a Web server and returned as an HTML list of
anchors to documents containing matching words.  This C code interfaces with a
library of functions provided especially for building such applications,
described in a document entitled "Swish-2.1.X search library".

SWISH-E 2.n is a version of the indexer package SWISH-E built by Kevin Hughes.
The original URL is http://sunsite.berkeley.edu/SWISH_E.  There, you can find
the documentation, mail list, etc.

This application has been developed using the VMS port of SWISH-E, by
Jean-Franois Pironne (jfp@altavista.net).


SUITABILITY
-----------
(Hopefully just about) any VMS Web server.  WASD CGILIB should make this script
suitable for WASD (all versions), OSU (all versions), CSWS (VMS Apache)
(versions 1.0-1 and later), Purveyor (all versions), possibly others.


USAGE
-----
Supply the search words either as an <IS-INDEX> style query string
("?first-word+second-word+last-word") or as form-encoded query string, where
the form field "words" supplies the search expression.  The form field "in" may
specify one or more (comma separated) keywords for confining the search to some
portion of an HTML document (e.g. TITLE, HEAD, BODY, etc.).

The index file must be supplied either as the request path (and therefore
available via CGI variable PATH_TRANSLATED, or in a form field named "index",
which must supply a path that can be translated by the server (performed by
redirection), or via the /INDEX= command-line qualifier which provides a file
specification acceptable to the C-RTL.

NOTE THAT THE PATH (either path or form field "index") MUST BE A SERVER-MAPPED
PATH THAT RESULTS IN THE SERVER PROVIDING A 'PATH_TRANSLATED' CGI VARIABLE!
It is not a native file specification.  This means all accesses to indices are
administered by the HTTP server.

An example form (with some bells-and-whistles):

  <HTML>
  <HEAD>
  <TITLE>Example SWISHesi Search Form</TITLE>
  </HEAD>
  <BODY>
  <FONT SIZE=+1><B>Example SWISH<SUP>esi</SUP> Search Form</B></FONT>
  <P>
  <FORM METHOD=GET ACTION="/cgi-bin/swishesi/ht_root/src/swish/swishesi/doc.index">
  Search Expression: <INPUT TYPE=text NAME=words SIZE=30><BR>
  Search Constraints: <SELECT NAME=in>
  <OPTION>title
  <OPTION>head
  <OPTION>body
  <OPTION>head,body
  <OPTION>title,head,body
  <OPTION>comments
  <OPTION SELECTED>all
  </SELECT><BR>
  Items-per-Page: <SELECT NAME=page>
  <OPTION>5
  <OPTION SELECTED>10
  <OPTION>15
  <OPTION>20
  <OPTION>50
  <OPTION>100
  </SELECT><BR>
  <INPUT TYPE=radio NAME=result VALUE="" CHECKED><I>default result page</I><BR>
  <INPUT TYPE=radio NAME=result VALUE="index_doc_result.html"><I>custom result page</I><BR>
  <INPUT TYPE=submit VALUE="Begin Search">
  <INPUT TYPE=reset VALUE="Reset">
  </FORM>
  </BODY>
  </HTML>

SWISHesi also allows "primary" and "alternate" indices (the index supplied
with the path is more efficient because it requires no redirect).  Provide a
selection list where the primary value is empty and the alternates are not.

  <FORM METHOD=GET ACTION="/cgi-bin/swishesi/web/index/one.swish-e">
  <INPUT TYPE=text SIZE=35 NAME=words><BR>
  <SELECT NAME=index>
  <OPTION VALUE="" SELECTED>one
  <OPTION VALUE="/web/index/two.swish-e">two
  <OPTION VALUE="/web/index/three.swish-e">three
  <OPTION VALUE="/web/index/four.swish-e">four
  </SELECT><BR>
  <INPUT TYPE=submit VALUE="Begin Search">
  </FORM>


CUSTOMIZING RESULTS PAGE
------------------------
Within some constraints SWISHesi provides for relatively easy customization of
the search result page.  This will allow for site-specific colour scheme,
banners, notes, etc., etc., to be introduced.  Basically SWISHesi interprets a
template result page that contains embedded, SSI-like directives, such as
<!--#RESULTS--> to add the table of search results.  Other HTML markup and text
is just output as part of the page.  THE FOLLOWING CAN BE USED AS A TEMPLATE IF
REQUIRED, just cut-and-paste.

  <BODY BGCOLOR="#ffffff" LINK="#0000ff" VLINK="#0000ff">
  <TABLE CELLPADDING=0 CELLSPACING=3 BORDER=0 WIDTH=100%>
  <TR>
  <TD ALIGN=left VALIGN=top><FONT SIZE=+1>
  <B><!--#DESCRIPTION--><SUP>&nbsp;</SUP></B>
  </FONT></TD>
  <TD ALIGN=right VALIGN=top><FONT SIZE=+1>
  <B>SWISH<SUP><U>esi</U></SUP></B>
  </FONT></TD>
  </TR>
  <TR><TD ALIGN=left><FONT SIZE=-1>&nbsp;<I><!--#ADMIN--></I></FONT></TD></TR>
  <!--#RESULTS-->
  <TR><TD>
  <B><!--#HITS-->,
  <!--#PAGES-->searching for &quot;<!--#WORD-->&quot;.</B>
  </TD></TR>
  </TABLE>
  </BODY>

By adding text and other directives the content and layout of the result page
can be changed to suit the application.  In this one the file link is being
suppressed, the last modified date/time no longer includes the weekday, page
backgound colour has been modified, the results description has been set to
some fixed text, search duration included, and a home page link has been added
following the next/previous "buttons".  Note that only the <BODY></BODY>
portion of the page can be customized.  SWISHesi supplies the <HEAD></HEAD>
portion.

  <!--#SET=FILE=0-->
  <!--#SET=MODIFIED=%d %b %Y %R-->
  <!--#SET=SUMMARY=<FONT SIZE=-1>%s</FONT>-->
  <BODY BGCOLOR="#99ccff" LINK="#0000ff" VLINK="#0000cc">
  <TABLE CELLPADDING=0 CELLSPACING=3 BORDER=0 WIDTH=100%>
  <TR>
  <TD ALIGN=left VALIGN=top><FONT SIZE=+1>
  <B>SWISH-E Index of WASD Documentation<SUP>&nbsp;</SUP></B>
  </FONT></TD>
  <TD ALIGN=right VALIGN=top><FONT SIZE=+1 COLOR="#0000ff>
  <B>SWISH<SUP><U>esi</U></SUP></B>
  </FONT></TD>
  </TR>
  <TR><TD ALIGN=left><FONT SIZE=-1>&nbsp;<I><!--#ADMIN--></I></FONT></TD></TR>
  <!--#RESULTS-->
  <TR><TD>
  <!--#HITS-->,
  <!--#PAGES-->searching for &quot;<!--#WORD-->&quot;.</B>
  &nbsp;(duration <!--#DURATION--> seconds)
  <BR>&nbsp;<B>[<A HREF="/">Site Home Page</A>]</B>
  </TD></TR>
  </TABLE>
  </BODY>

The directives supported are:

<!--#ADMIN-->                  include index admin here (IndexAdmin)
<!--#DURATION-->               include search duration here (floating seconds)
<!--#DESCRIPTION-->            include index descrip here (IndexDescription)
<!--#HITS-->                   number of hits followed by "hits[s]"
<!--#IN-->                     what was being searched ("FILE", "TITLE", etc.)
<!--#PAGES-->                  include page "buttons" here
<!--#RESULTS-->                include results table here
<!--#SET=FILE=integer-->       zero suppresses file link, non-zero includes
<!--#SET=MODIFIED=string-->    set the strftime() format string
<!--#SET=SUMMARY=integer-->    set the maximum length of file summary
<!--#SET=SUMMARY=string-->     printf() format string around file summary
                               (e.g. "<FONT SIZE=-1><I>%s</I></FONT>")
                               MUST have only one "%s"!
<!--#UPDATED[=string]-->       index last updated, optional strftime() format
<!--#WORD-->                   the string(s) being search for

There is an internal, default result page template.  When a custom result page
is required it may be specified using the /RESULT=full-file-specification
command-line qualifiers or by using the RESULT form field, as follows

  <INPUT TYPE=hidden NAME=result VALUE="web.html">

where "web.html" is the result page template located in the same directory as
the specified index file.


CGIPLUS
-------
For the WASD CGIplus persistant scripting environment this script buffers the
last requested search, index file name and search expressions (words) between
uses.  When another request is received it compares the name of the index file
and search expression and with that buffered from the previous request.  If
identical the previous search results is used again to supply the requested
results.  This can be particularly for clients just stepping through a series
of ranges of search results.  If not identical the previous search is closed
and a new one initiated.  Let's hope SWISH-E doesn't leak too much!


CGI ENVIRONMENT VARIABLES
-------------------------
FORM_HIT              begin at hit number (one by default)
FORM_INDEX            with path to index (will redirect to map)
FORM_IN               "FILE", "TITLE", "HEAD", "BODY", "COMMENTS",
                      "HEADER", "EMPHASIZED", the 'structure' of SwishSearch()
FORM_PAGE             number of items per page
FORM_RESULT           result page template (must be located with index file)
FORM_SUMMARY          length of any summary included (0 suppresses)
FORM_WIN              number of hits per page (25 by default)
FORM_WORD[S]          contains search words (-w)
KEY_n                 space-separated search words (-w)
PATH_INFO             path to the SWISH index
PATH_TRANSLATED       file specification of the SWISH index (-f)


QUALIFIERS
----------
/DBUG                 turns on all "if (Debug)" statements
/RESULT=file          search result page template (can be located anywhere)
/VERSION              display the SWISHesi and CGILIB versions


LOGICAL/SYMBOL NAMES
--------------------
SWISHESI$DBUG         turns on all "if (Debug)" statements
SWISHESI$PARAM        used as if it was the command-line (parameters)
                      (OSU users require P1, P2, P3 as '"" "" ""')


BUILD DETAILS
-------------
Language-specific messages are available.  See "lang_en.h".
See BUILD_SWISHESI.COM procedure.


COPYRIGHT
---------
Copyright (C) 2000-2002 Mark G.Daniel (mark.daniel@wasd.vsm.com.au)
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.

SWISH-E Copyright (C) 1995,1996,1997,1998 Hewlett-Packard Company
Originally by Kevin Hughes (kev@kevcom.com), 3/11/1994
This program and library is free software; you can redistribute it and/or
modify it under the terms of the GNU (Library) General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

VMS port of SWISH-E, by Jean-Franois Pironne (jfp@altavista.net).


VERSION HISTORY (update SOFTWAREID as well!)
---------------
05-JAN-2002  JFP  v0.4.7, minor adjustments for dev-25
09-SEP-2001  MGD  v0.4.6, continuing development
07-JUL-2001  MGD  v0.4.5, continuing development
25-MAY-2001  MGD  v0.3.0, relauched upon an unsuspecting VMS community
25-NOV-2000  MGD  v0.2.0, initial development, SWISH-E v2.1.8
*/
/*****************************************************************************/

#ifdef __ALPHA
#  define SOFTWAREID "SWISHESI AXP-0.4.7"
#else
#  define SOFTWAREID "SWISHESI VAX-0.4.7"
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* VMS-related header files */
#include <descrip.h>
#include <ssdef.h>
#include <libdtdef.h>
#include <lib$routines.h>
#include <stsdef.h>

/* application header files */
#include "cgilib.h"
#include "lang_en.h"
#include "swishe_root:[src]swish.h"
#include "swishe_root:[src]error.h"
#include "swishe_root:[src]list.h"
#include "swishe_root:[src]docprop.h"
#include "swishe_root:[src]result_output.h"
#include "swishe_root:[src]mem.h"
#include "swishe_root:[src]search.h"
#undef free

#define boolean int
#define true 1
#define false 0

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))
 
#define FI_LI __FILE__,__LINE__

#define DEFAULT_PAGE_SIZE 10
#define DEFAULT_SUMMARY_FORMAT "<FONT SIZE=-1>%s</FONT>"
#define DEFAULT_SUMMARY_SIZE 9999
#define DEFAULT_INCLUDE_FILE_LINK 1
#define DEFAULT_LAST_MODIFIED_FORMAT "%a, %d %b %Y %R"
#define DEFAULT_UPDATED_FORMAT "%a, %d %b %Y %R"

#define NOTHING_MATCHING 0

char  DefaultResultPage [] =
"<BODY BGCOLOR=\"#ffffff\" LINK=\"#0000ff\" VLINK=\"#0000ff\">\n\
<TABLE CELLPADDING=0 CELLSPACING=3 BORDER=0 WIDTH=100%>\n\
<TR>\n\
<TD ALIGN=left VALIGN=top><FONT SIZE=+1>\n\
<B><!--#DESCRIPTION--><SUP>&nbsp;</SUP></B>\n\
</FONT></TD>\n\
<TD ALIGN=right VALIGN=top><FONT SIZE=+1>\n\
<B>SWISH<SUP><U>esi</U></SUP></B>\n\
</FONT></TD>\n\
</TR>\n\
<TR><TD ALIGN=left><FONT SIZE=-1>&nbsp;\n\
<I><!--#ADMIN--></I>\n\
</FONT></TD></TR>\n\
<TR><TD>\n\
<!--#RESULTS-->\n\
<TR><TD>\n\
<B><!--#HITS-->,\n\
<!--#PAGES-->searching for &quot;<!--#WORD-->&quot;.</B>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n";

char  CopyrightInfo [] =
"Copyright (C) 2000,2001 Mark G.Daniel.\n\
This software comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome to redistribute it\n\
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.";

char  Utility [] = "SWISHESI";

boolean  Debug,
         IsCgiPlus,
         NoCache;

int  BeginAtHit,
     CgiPlusUsageCount,
     EndAtHit,
     PageSize,
     SummarySize,
     ResultPageLength;

char  SoftwareID [128],
      SearchWords [256];

char  *CgiEnvironmentPtr,
      *CgiFormInPtr,
      *CgiFormModifiedPtr,
      *CgiFormResultPtr,
      *CgiKey1Ptr,
      *CgiPathInfoPtr,
      *CgiQueryStringPtr,
      *CgiScriptNamePtr,
      *CliIndexFileNamePtr,
      *CliResultPagePtr,
      *DescriptionPtr,
      *IndexFileNamePtr,
      *ResultPagePtr;

/* required prototypes */
int GetParameters ();
void ProcessRequest ();
int ReadFileIntoMemory (char*, char**, int*);
char *SwishEsiError (int);
void SwishEsiSearch (char*, char*, char*);
char* SwishHeaderParameter (IndexFILE*, char*);
int strsame (char*, char*, int);
int sys$gettim (void*);

/*****************************************************************************/
/*
*/

main (int argc, char *argv[])

{
   int  KeyCount;
   char  *cptr, *sptr, *zptr,
         *SearchWordsPtr;
   char  KeyName [32];

   /*********/
   /* begin */
   /*********/

   sprintf (SoftwareID, "%s (%s) (SWISH-E %s)",
            SOFTWAREID, CGILIB_SOFTWAREID, SWISH_VERSION);

   GetParameters ();

   if (CliResultPagePtr != NULL && !CliResultPagePtr[0])
   {
      /* output the default results page (can be used as a template) */
      fputs (DefaultResultPage, stdout);
      exit (SS$_NORMAL);
   }

   if (getenv ("SWISHESI$DBUG") != NULL) Debug = true;
   CgiLibEnvironmentSetDebug (Debug);

   CgiLibEnvironmentInit (argc, argv, false);
   CgiEnvironmentPtr = CgiLibEnvironmentName ();
   IsCgiPlus = CgiLibEnvironmentIsCgiPlus ();

   CgiLibResponseSetSoftwareID (SoftwareID);
   CgiLibResponseSetErrorStatus (404);
   CgiLibResponseSetErrorMessage (LANG_ReportedBySwishEsi);

   if (IsCgiPlus)
   {
      for (;;)
      {
         /* block waiting for the next request */
         CgiLibVar ("");
         CgiPlusUsageCount++;
         ProcessRequest ();
         CgiLibCgiPlusEOF ();
      }
   }
   else
      ProcessRequest ();

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

void ProcessRequest ()

{
   int  status,
        KeyCount;
   char  *cptr, *sptr, *zptr,
         *SearchWordsPtr;
   char  KeyName [32],
         ResultPageFile [256];

   /*********/
   /* begin */
   /*********/

   CgiScriptNamePtr = CgiLibVar ("WWW_SCRIPT_NAME");

   if (IsCgiPlus)
   {
      cptr = CgiLibVar ("WWW_HTTP_PRAGMA");
      if (strsame (cptr, "no-cache", -1))
         NoCache = true;
      else
         NoCache = false;
   }

   BeginAtHit = 1;
   PageSize = DEFAULT_PAGE_SIZE;
   CgiFormInPtr = NULL;

   if ((cptr = CgiLibVarNull ("WWW_KEY_1")) != NULL)
   {
      /******************************/
      /* key-words ("?one+two+etc") */
      /******************************/

      CgiPathInfoPtr = CgiLibVar ("WWW_PATH_INFO");
      if (CliIndexFileNamePtr == NULL)
         IndexFileNamePtr = CgiLibVar ("WWW_PATH_TRANSLATED");
      else
         IndexFileNamePtr = CliIndexFileNamePtr;
      zptr = (sptr = SearchWordsPtr = SearchWords) + sizeof(SearchWords);
      for (KeyCount = 1; KeyCount < 100; KeyCount++)
      {
         sprintf (KeyName, "WWW_KEY_%d", KeyCount);
         if ((cptr = CgiLibVarNull (KeyName)) == NULL) break;
         if (sptr > SearchWords && sptr < zptr) *sptr++ = ' ';
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      }
      *sptr = '\0';
   }
   else
   {
      /***************/
      /* form fields */
      /***************/

      CgiPathInfoPtr = CgiLibVar ("WWW_PATH_INFO");
      if (CliIndexFileNamePtr == NULL)
      {
         IndexFileNamePtr = CgiLibVar ("WWW_PATH_TRANSLATED");
         if (!IndexFileNamePtr[0])
         {
            /* redirect to get index file form field into request path */
            IndexFileNamePtr = CgiLibVar ("FORM_INDEX");
            if (!IndexFileNamePtr[0])
            {
               CgiLibResponseError (FI_LI, 0, LANG_IndexFileNotSpecified);
               return;
            }
            CgiQueryStringPtr = CgiLibVar ("WWW_QUERY_STRING");
            /* prevent redirection "loops" (OSU exhibits this) */
            if ((cptr = strstr (CgiQueryStringPtr, "index=")) != NULL)
               memcpy (cptr, "xxxxx", 5);
            CgiLibResponseRedirect ("%s%s?%s", CgiScriptNamePtr,
                                    IndexFileNamePtr, CgiQueryStringPtr);
            return;
         }
      }
      else
         IndexFileNamePtr = CliIndexFileNamePtr;

      CgiFormResultPtr = "";
      if ((cptr = CliResultPagePtr) == NULL)
         if (*(CgiFormResultPtr = CgiLibVar("WWW_FORM_RESULT")))
         {
            zptr = (sptr = ResultPageFile) + sizeof(ResultPageFile)-1;
            for (cptr = IndexFileNamePtr;
                 *cptr && sptr < zptr;
                 *sptr++ = *cptr++);
            *sptr = '\0';
            while (sptr > ResultPageFile && *sptr != ']') sptr--;
            while (*sptr == ']') sptr++;
            cptr = CgiFormResultPtr;
            while (*cptr && sptr < zptr) *sptr++ = *cptr++;
            *sptr = '\0';
            cptr = ResultPageFile;
         }

      if (cptr == NULL || !cptr[0])
         ResultPagePtr = DefaultResultPage;
      else
      {
         status = ReadFileIntoMemory (cptr, &ResultPagePtr, &ResultPageLength);
         if (VMSnok (status))
         {
            CgiLibResponseError (FI_LI, status, cptr);
            exit (SS$_NORMAL);
         }
      }

      /* search CGI variables */

      SearchWordsPtr = CgiLibVar ("FORM_WORD");
      if (!SearchWordsPtr[0]) SearchWordsPtr = CgiLibVar ("FORM_WORDS");

      cptr = CgiLibVar ("WWW_FORM_HIT");
      if (cptr[0]) BeginAtHit = atoi (cptr);

      cptr = CgiLibVar ("WWW_FORM_PAGE");
      if (cptr[0]) PageSize = atoi (cptr);

      CgiFormInPtr = CgiLibVar ("WWW_FORM_IN");
   }
      
   SwishEsiSearch (IndexFileNamePtr, SearchWordsPtr, CgiFormInPtr);

   if (ResultPageLength)
   {
      /* must have been read from a file, free the allocated memory */
      free (ResultPagePtr);
      ResultPageLength = 0;
   }
}

/****************************************************************************/
/*
Using the SWISH-E search API perform a search of the specific index.  Then
using the result page template build and output a page including the search
results.
*/ 

void SwishEsiSearch 
(
char *IndexFileName,
char *SearchWords,
char *SearchIn
)
{
   static int  LibDeltaSecondsF = LIB$K_DELTA_SECONDS_F;
   static int  PrevSearchStructure,
               SwishStatus,
               SearchStructure;
   static char  PrevIndexFileName [256],
                PrevSearchWords [256];
   static struct IndexFILE *IndexListPtr;
   static RESULT  *ResultPtr;
   static SWISH  *SwishPtr;
   static int ResultsCount;

   boolean  IncludeFileLink;
   int  cnt,
        HitCount,
        HitRank,
        PageEnd,
        PageStart,
        PageStartWindow,
        PageSizeSquared;
   unsigned long  SearchBinTime [2],
                  StartBinTime [2],
                  EndBinTime [2];
   float  SearchSeconds;
   char  *cptr, *sptr, *zptr,
         *LastModifiedFormatPtr,
         *SummaryFormatPtr,
         *UpdatedFormatPtr;
   char  LastModified [64],
         LastModifiedFormat [128],
         StringHtmlEsc [256],
         SearchWordsUrlEnc [256],
         SizeString [16],
         SummaryFormat [128],
         SummaryBuffer [1024],
         UntitledBuffer [128],
         UpdatedFormat [128],
         UpdatedTime [64];
   struct tm  *tmptr;
   stat_t  StatBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SwishEsiSearch() |%s|%s|\n",
               IndexFileName, SearchWords);

   SearchStructure = IN_FILE;
   if (SearchIn != NULL && SearchIn[0])
   {
      SearchStructure = 0;
      cptr = SearchIn; 
      while (*cptr)
      {
         while (*cptr && (isspace(*cptr) || ispunct(*cptr))) cptr++;
         if (strsame (cptr, "file", 4))
            SearchStructure |= IN_FILE;
         else
         if (strsame (cptr, "title", 5))
            SearchStructure |= IN_TITLE;
         else
         if (strsame (cptr, "head", 4))
            SearchStructure |= IN_HEAD;
         else
         if (strsame (cptr, "body", 4))
            SearchStructure |= IN_BODY;
         else
         if (strsame (cptr, "comments", 8))
            SearchStructure |= IN_COMMENTS;
         else
         if (strsame (cptr, "header", 6))
            SearchStructure |= IN_HEADER;
         else
         if (strsame (cptr, "emphasized", 10))
            SearchStructure |= IN_EMPHASIZED;
         else
         if (strsame (cptr, "all", 3))
            SearchStructure |= IN_ALL;
         else
         {
            CgiLibResponseError (FI_LI, 0, LANG_UnknownIn);
            return;
         }
         while (*cptr && !isspace(*cptr) && !ispunct(*cptr)) cptr++;
      }
   }

   if (IsCgiPlus && PrevIndexFileName[0])
   {
      if (NoCache ||
          !strsame (IndexFileName, PrevIndexFileName, -1) ||
          !strsame (SearchWords, PrevSearchWords, -1) ||
          SearchStructure != PrevSearchStructure)
      {
         SwishClose (SwishPtr);
         SwishPtr = NULL;
         PrevIndexFileName[0] = PrevSearchWords[0] = '\0';
         PrevSearchStructure = 0;
      }
   }

   if (SwishPtr == NULL)
   {
      int  desc = 1;
      int  asc = -1;

      /* Start a session */
      SwishPtr = SwishNew();

      /* set absolute minimum search verbosity */
      SwishPtr->ResultOutput->headerOutVerbose = 0;

      /*
         03-JUL-2001 MGD
         This stat() to determine whether an index file exists is a kludge!
         BETA addindexfile() insists on providing a plain-text <stdout> error
         message if it can't get to the index file.  No good for CGI scripts.
         Can be removed sometime in the future if this behaviour changes. 
      */
      if (stat (IndexFileName, &StatBuffer) < 0)
      {
         CgiLibResponseError (FI_LI, 0, LANG_SwishIndexNotFound);
         return;
      }

      /* Set default index file, if none specified */
      if (!SwishPtr->indexlist)
         SwishPtr->indexlist = addindexfile(SwishPtr->indexlist, IndexFileName);

      addSearchResultSortProperty(SwishPtr, AUTOPROPERTY_RESULT_RANK, desc);
      addSearchResultSortProperty(SwishPtr, AUTOPROPERTY_DOCPATH, asc);
      addSearchResultSortProperty(SwishPtr, AUTOPROPERTY_TITLE, asc);
      addSearchResultSortProperty(SwishPtr, AUTOPROPERTY_DOCSIZE, asc);
      addSearchResultSortProperty(SwishPtr, AUTOPROPERTY_LASTMODIFIED, asc);
      addSearchResultSortProperty(SwishPtr, AUTOPROPERTY_SUMMARY, asc);

      SwishStatus = SwishAttach (SwishPtr, 0);
      if (SwishStatus < 0)
      {
         CgiLibResponseError (FI_LI, 0, SwishEsiError(SwishStatus));
         return;
      }

      IndexListPtr = SwishPtr->indexlist;

      sys$gettim (&StartBinTime);

      ResultsCount = SwishStatus = search (SwishPtr, SearchWords, SearchStructure);
      if (SwishStatus < 0)
      {
         CgiLibResponseError (FI_LI, 0, SwishEsiError(SwishStatus));
         return;
      }

      sys$gettim (&EndBinTime);
   }

   SwishStatus = ResultsCount;
   if (SwishStatus < 0)
      CgiLibResponseError (FI_LI, 0, SwishEsiError(SwishStatus));
   else
#if NOTHING_MATCHING
   if (SwishStatus == 0)
      CgiLibResponseError (FI_LI, 0, LANG_NothingMatching);
   else
#endif
   {
      lib$sub_times (&EndBinTime, &StartBinTime, &SearchBinTime);
      lib$cvtf_from_internal_time (&LibDeltaSecondsF, &SearchSeconds,
                                   &SearchBinTime);

      if (SwishStatus && BeginAtHit > SwishStatus)
      {
         CgiLibResponseError (FI_LI, 0, LANG_NotThatManyHits);
         return;
      }
      if ((EndAtHit = BeginAtHit + PageSize - 1) <= 0)
      {
         CgiLibResponseError (FI_LI, 0, LANG_InvalidPageSize);
         return;
      }

      IncludeFileLink = DEFAULT_INCLUDE_FILE_LINK;
      LastModifiedFormatPtr = DEFAULT_LAST_MODIFIED_FORMAT;
      SummaryFormatPtr = DEFAULT_SUMMARY_FORMAT;
      SummarySize = DEFAULT_SUMMARY_SIZE;

      /**********************/
      /* result page <HEAD> */
      /**********************/

      CgiLibResponseHeader (200, "text/html");

      fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"duration\" CONTENT=\"%.3f Seconds\">\n\
<TITLE>SWISHesi</TITLE>\n\
</HEAD>\n",
         SoftwareID, 
         CgiEnvironmentPtr,
         SearchSeconds);

      /**********************/
      /* result page <BODY> */
      /**********************/

      while (*ResultPagePtr)
      {
         /* look for SSI-like directives embedded in the page HTML */
         for (cptr = ResultPagePtr; *cptr; cptr++)
            if (*cptr == '<' && !memcmp(cptr, "<!--#", 5)) break;
         /* output any intervening characters */
         if (cptr - ResultPagePtr)
            fwrite (ResultPagePtr, cptr - ResultPagePtr, 1, stdout);
         /* if got to the end-of-page */
         if (!*cptr) break;

         if (strsame (cptr, "<!--#ADMIN-->", 13))
         {
            /*********/
            /* admin */
            /*********/

            ResultPagePtr = cptr + 13;
            /* "admin" can include HTML markup if desired */
            cptr = SwishHeaderParameter (IndexListPtr, "IndexAdmin");
            if (cptr != NULL && cptr[0]) fputs (cptr, stdout);
            continue;
         }

         if (strsame (cptr, "<!--#DESCRIPTION-->", 19))
         {
            /***************/
            /* description */
            /***************/

            ResultPagePtr = cptr + 19;
            cptr = SwishHeaderParameter (IndexListPtr, "Description");
            if (cptr == NULL || !cptr[0]) cptr = LANG_SearchResults;
            cptr = (char*)CgiLibHtmlEscape (cptr, -1, NULL, -1);
            fputs (cptr, stdout);
            free (cptr);
            continue;
         }

         if (strsame (cptr, "<!--#DURATION-->", 16))
         {
            /*******************/
            /* search duration */
            /*******************/

            ResultPagePtr = cptr + 16;
            fprintf (stdout, "%.3f", SearchSeconds);
            continue;
         }

         if (strsame (cptr, "<!--#HITS-->", 12))
         {
            /******************/
            /* number of hits */
            /******************/

            ResultPagePtr = cptr + 12;
            if (SwishStatus == 1)
               fprintf (stdout, "%d %s", SwishStatus, LANG_Hit);
            else
               fprintf (stdout, "%d %s", SwishStatus, LANG_Hits);
            continue;
         }

         if (strsame (cptr, "<!--#IN-->", 10))
         {
            /************/
            /* in what? */
            /************/

            ResultPagePtr = cptr + 10;
            if (SearchIn != NULL && SearchIn[0])
            {
               CgiLibHtmlEscape (SearchIn, -1,
                                 StringHtmlEsc, sizeof(StringHtmlEsc));
               fputs (StringHtmlEsc, stdout);
            }
            else
               fputs ("?", stdout);
            continue;
         }

         /* "#next" only for backward-compatibility */
         if (strsame (cptr, "<!--#PAGES-->", 13) ||
             strsame (cptr, "<!--#NEXT-->", 12))
         {
            /************************/
            /* page range "buttons" */
            /************************/

            ResultPagePtr = cptr + 13;
            if (toupper(cptr[5]) == 'N') ResultPagePtr--;

            /* if it all fits on a single page then forget this */
            if (SwishStatus <= PageSize) continue;

            CgiLibUrlEncode (SearchWords, -1,
                             SearchWordsUrlEnc, sizeof(SearchWordsUrlEnc));

            PageSizeSquared = PageSize * PageSize;
            PageStart = 1;
            PageEnd = PageStart + PageSizeSquared - 1;
            /* big page ranges prior to current page */
            while (PageEnd < BeginAtHit)
            {
               fprintf (stdout,
"[<A HREF=\"%s%s?word=%s&in=%s&hit=%d&page=%d&result=%s\">%d-%d</A>]\n",
                  CgiScriptNamePtr, CgiPathInfoPtr,
                  SearchWordsUrlEnc, SearchIn, PageStart,
                  PageSize, CgiFormResultPtr, PageStart, PageEnd);
               PageStart += PageSizeSquared;
               PageEnd = PageStart + PageSizeSquared - 1;
            }
            PageStartWindow = PageStart;
            PageEnd = PageStart + PageSize - 1;
            /* page-size ranges prior to current page */
            while (PageEnd < BeginAtHit)
            {
               fprintf (stdout,
"[<A HREF=\"%s%s?word=%s&in=%s&hit=%d&page=%d&result=%s\">%d-%d</A>]\n",
                  CgiScriptNamePtr, CgiPathInfoPtr,
                  SearchWordsUrlEnc, SearchIn, PageStart,
                  PageSize, CgiFormResultPtr, PageStart, PageEnd);
               PageStart += PageSize;
               PageEnd = PageStart + PageSize - 1;
            }
            /* current page */
            if (PageStart == SwishStatus)
               fprintf (stdout, "[%d]\n", PageStart);
            else
            if (PageEnd < SwishStatus)
               fprintf (stdout, "[%d-%d]\n", PageStart, PageEnd);
            else
               fprintf (stdout, "[%d-%d]\n", PageStart, SwishStatus);
            PageStart += PageSize;
            PageEnd = PageStart + PageSize - 1;
            /* page-size ranges after the current page */
            while (PageStart <= SwishStatus &&
                   PageEnd < PageStartWindow + PageSizeSquared)
            {
               fprintf (stdout,
"[<A HREF=\"%s%s?word=%s&in=%s&hit=%d&page=%d&result=%s\">",
                  CgiScriptNamePtr, CgiPathInfoPtr,
                  SearchWordsUrlEnc, SearchIn, PageStart,
                  PageSize, CgiFormResultPtr);
               if (PageStart == SwishStatus)
                  fprintf (stdout, "%d</A>]\n", PageStart);
               else
               if (PageEnd < SwishStatus)
                  fprintf (stdout, "%d-%d</A>]\n", PageStart, PageEnd);
               else
                  fprintf (stdout, "%d-%d</A>]\n", PageStart, SwishStatus);
               PageStart += PageSize;
               PageEnd = PageStart + PageSize - 1;
            }
            /* big page ranges after current page */
            while (PageStart < SwishStatus)
            {
               PageEnd = PageStart + PageSizeSquared - 1;
               fprintf (stdout,
"[<A HREF=\"%s%s?word=%s&in=%s&hit=%d&page=%d&result=%s\">",
                  CgiScriptNamePtr, CgiPathInfoPtr,
                  SearchWordsUrlEnc, SearchIn, PageStart,
                  PageSize, CgiFormResultPtr);
               if (PageStart == SwishStatus)
                  fprintf (stdout, "%d</A>]\n", PageStart);
               else
               if (PageEnd < SwishStatus)
                  fprintf (stdout, "%d-%d</A>]\n", PageStart, PageEnd);
               else
                  fprintf (stdout, "%d-%d</A>]\n", PageStart, SwishStatus);
               PageStart += PageSizeSquared;
               PageEnd = PageStart + PageSizeSquared - 1;
            }

            continue;
         }

         if (strsame (cptr, "<!--#RESULTS-->", 15))
         {
            /***********/
            /* results */
            /***********/

            ResultPagePtr = cptr + 15;

            /* zero hits then no need for results table */
            if (!SwishStatus) continue;

            fprintf (stdout,
"<TABLE CELLPADDING=0 CELLSPACING=3 BORDER=0>\n\
<TR>\
<TH></TH>\
<TH ALIGN=left><NOBR><U>%s</U>&nbsp;&nbsp;</NOBR></TH>\
<TH ALIGN=right><NOBR><U>%s</U>&nbsp;&nbsp;</NOBR></TH>\
<TH ALIGN=right><NOBR><U>%s</U></NOBR></TH>\
</TR>\n\
<TR HEIGHT=3></TR>\n",
               LANG_Title, LANG_LastModified, LANG_Size);

            /* for both CGIplus and general begin-at-more-than-one */

            SwishSeek (SwishPtr, BeginAtHit - 1);
            HitCount = BeginAtHit - 1;

            while ((ResultPtr = SwishNext (SwishPtr)) != NULL)
            {
               PropValue* pvDocPath;
               PropValue* pvTitle;
               PropValue* pvDocSize;
               PropValue* pvLastModified;
               PropValue* pvSummary;
               char* filename;
               char* title;
               time_t last_modified;
               char* summary;
               long size;
              
               pvDocPath = getResultPropValue(SwishPtr, ResultPtr, 
                                              AUTOPROPERTY_DOCPATH, 0);
               filename = pvDocPath->value.v_str;
               pvTitle = getResultPropValue(SwishPtr, ResultPtr, 
                                              AUTOPROPERTY_TITLE, 0);
               title = pvTitle->value.v_str;
               pvDocSize = getResultPropValue(SwishPtr, ResultPtr, 
                                              AUTOPROPERTY_DOCSIZE, 0);
               size = pvDocSize->value.v_ulong;
               pvLastModified = getResultPropValue(SwishPtr, ResultPtr, 
                                              AUTOPROPERTY_LASTMODIFIED, 0);
               last_modified = pvLastModified->value.v_date;
               pvSummary = getResultPropValue(SwishPtr, ResultPtr, 
                                              AUTOPROPERTY_SUMMARY, 0);
               summary = pvSummary->value.v_str;

               if (Debug)
                 fprintf (stdout, "%d %s \"%s\" %d\n",
                          ResultPtr->rank, filename,
                          title, size);

               /* CGIplus, keep the ranks being 'overly' adjusted */
               HitRank = ResultPtr->rank;
               HitCount++;
               if (HitCount < BeginAtHit) continue;
               if (HitCount > EndAtHit) break;

               if (!*title)
               {
                  /* not title (plain text?), use file "name" */
                  for (cptr = filename; *cptr; cptr++);
                  while (cptr > filename && *cptr != '/') cptr--;
                  if (*cptr == '/') cptr++;
                  zptr = (sptr = UntitledBuffer) + sizeof(UntitledBuffer)-1;
                  *sptr++ = '/';
                  while (*cptr && sptr < zptr) *sptr++ = *cptr++;
                  *sptr = '\0';
                  title = UntitledBuffer;
               }

               if (size > 1000000)
                 sprintf (SizeString, "%dMB", size / 1000000);
               else
               if (size > 1000)
                 sprintf (SizeString, "%dkB", size / 1000);
               else
                 sprintf (SizeString, "%dB", size);

               tmptr = localtime (&last_modified);
               if (!strftime (LastModified, sizeof(LastModified),
                              LastModifiedFormatPtr, tmptr))
                  strcpy (LastModified, "strftime() error!");

               fprintf (stdout,
"<TR>\
<TH ALIGN=right VALIGN=top>%d.&nbsp;&nbsp;</TH>\
<TD ALIGN=left VALIGN=top><A HREF=\"%s\">%s</A>&nbsp;&nbsp;</TD>\
<TD ALIGN=right VALIGN=top><NOBR>%s&nbsp;&nbsp;</NOBR></TD>\
<TD ALIGN=right VALIGN=top><NOBR>%s</NOBR></TD>\
</TR>\n\
<!-- rank=%d -->\n\
",
                  HitCount,
                  filename,
                  title,
                  LastModified,
                  SizeString, 
                  HitRank);

               if (SummarySize && summary != NULL)
               {
                  /* trim up the (potentially) ragged ends of the summary */
                  for (cptr = summary; *cptr; cptr++)
                  {
                     if (*cptr == '&' && !memcmp(cptr, "&nbsp;", 6))
                        cptr += 6;
                     else
                     if (!isspace(*cptr))
                        break;
                  }
                  zptr = (sptr = SummaryBuffer) + sizeof(SummaryBuffer);
                  for (cnt = SummarySize; cnt && *cptr && sptr < zptr; cnt--)
                     *sptr++ = *cptr++;
                  *sptr = '\0';
                  while (sptr > SummaryBuffer && !isspace(*sptr)) sptr--;
                  if (sptr == SummaryBuffer)
                     for (cnt = SummarySize; cnt && *sptr; cnt--) sptr++;
                  *sptr = '\0';
                  fputs ("<TR><TD></TD><TD COLSPAN=3 ALIGN=left>", stdout);
                  fprintf (stdout, SummaryFormatPtr, SummaryBuffer);
                  fputs ("</TD></TR>\n", stdout);
               }

               if (IncludeFileLink)
                  fprintf (stdout,
"<TR>\
<TD></TD>\
<TD COLSPAN=3 ALIGN=right><FONT SIZE=-1><A HREF=\"%s\">%s</A></FONT></TD>\
</TR>\n",
                     filename, filename);

               fputs ("<TR HEIGHT=3></TR>\n", stdout);

               efree(pvDocPath);
               efree(pvTitle);
	       efree(pvDocSize);
	       efree(pvLastModified);
	       efree(pvSummary);
            }

            fputs ("</TABLE>", stdout);
            continue;
         }

         /**********************/
         /* index last updated */
         /**********************/

         if (strsame (cptr, "<!--#UPDATED", 12))
         {
            cptr += 12;
            if (*cptr == '=') cptr++;
            zptr = (sptr = UpdatedFormat) + sizeof(UpdatedFormat)-1;
            while (*cptr && (*cptr != '-' || memcmp(cptr, "-->", 3)) &&
                   sptr < zptr) *sptr++ = *cptr++;
            *sptr = '\0';
            while (*cptr && (*cptr != '-' || memcmp(cptr, "-->", 3))) cptr++;
            if (*cptr) cptr += 3;
            ResultPagePtr = cptr;

            if (UpdatedFormat[0] && !isspace(UpdatedFormat[0]))
               UpdatedFormatPtr = UpdatedFormat;
            else
               UpdatedFormatPtr = DEFAULT_UPDATED_FORMAT;
            if (!stat (IndexFileName, &StatBuffer))
            {
               tmptr = localtime (&StatBuffer.st_mtime);
               if (strftime (UpdatedTime, sizeof(UpdatedTime),
                             UpdatedFormatPtr, tmptr))
                  fputs (UpdatedTime, stdout);
               else
                  fputs ("strftime() error!", stdout);
            }
            else
               fputs ("stat() error!", stdout);

            continue;
         }

         if (strsame (cptr, "<!--#WORD-->", 12) ||
             strsame (cptr, "<!--#WORDS-->", 13))
         {
            /************************/
            /* searched for word(s) */
            /************************/

            ResultPagePtr = cptr + 12;
            if (toupper(cptr[9]) == 'S') ResultPagePtr++;
            CgiLibHtmlEscape (SearchWords, -1,
                              StringHtmlEsc, sizeof(StringHtmlEsc));
            fputs (StringHtmlEsc, stdout);
            continue;
         }

         /************/
         /* SETtings */
         /************/

         if (strsame (cptr, "<!--#SET=FILE=", 14))
         {
            cptr += 14;
            IncludeFileLink = atoi(cptr);
         }
         else
         if (strsame (cptr, "<!--#SET=MODIFIED=", 18))
         {
            cptr += 18;
            LastModifiedFormatPtr = LastModifiedFormat;
            zptr = (sptr = LastModifiedFormat) + sizeof(LastModifiedFormat)-1;
            while (*cptr && (*cptr != '-' || memcmp(cptr, "-->", 3)) &&
                   sptr < zptr) *sptr++ = *cptr++;
            *sptr = '\0';
         }
         else
         if (strsame (cptr, "<!--#SET=SUMMARY=", 17))
         {
            cptr += 17;
            if (isdigit(*cptr))
               SummarySize = atoi(cptr);
            {
               zptr = (sptr = SummaryFormat) + sizeof(SummaryFormat)-1;
               while (*cptr && (*cptr != '-' || memcmp(cptr, "-->", 3)) &&
                      sptr < zptr) *sptr++ = *cptr++;
               *sptr = '\0';
               SummaryFormatPtr = SummaryFormat;
            }
         }
         else
            /* can't believe I'm doing this (post-Dikjstra) */
            goto ErrorUnknown;

         /* just scan to the end of SETting directive */
         while (*cptr && (*cptr != '-' || memcmp(cptr, "-->", 3))) cptr++;
         if (*cptr) cptr += 3;
         /* SETtings are often on a line alone, suppress trailing newline */
         if (*cptr == '\n') cptr++;
         ResultPagePtr = cptr;
         continue;

         /*********/
         /* error */
         /*********/

         ErrorUnknown:

         for (sptr = cptr; *sptr && *sptr != '>'; sptr++);
         if (*sptr) sptr++;
         cptr = (char*)CgiLibHtmlEscape (cptr, sptr-cptr, NULL, -1); 
         fprintf (stdout,
"<P><FONT SIZE=+1 COLOR=\"#ff000\"><B>ERROR: %s</B></FONT>\n\
</BODY>\n",
            cptr);
         free (cptr);
         break;
      }

      fputs ("</HTML>\n", stdout);
   }
      
   if (IsCgiPlus)
   {
      /* buffer this request's details in case they're can be reused */
      strncpy (PrevIndexFileName, IndexFileName, sizeof(PrevIndexFileName));
      PrevIndexFileName[sizeof(PrevIndexFileName)-1] = '\0';
      strncpy (PrevSearchWords, SearchWords, sizeof(PrevSearchWords));
      PrevSearchWords[sizeof(PrevSearchWords)-1] = '\0';
   }
   else
      SwishClose (SwishPtr);
}

/****************************************************************************/
/*
Read the file contents specified by 'FileName' into memory, set the pointer
at 'FileTextPtr' to the contents and the file size at 'FileSizePtr'.  Returns a
VMS status value that should be checked.
*/ 

int ReadFileIntoMemory
(
char *Source,
char **FileTextPtr,
int *FileSizePtr
)
{
   static int  BytesRemaining;

   int  status,
        Bytes,
        BufferCount,
        Length;
   char  *BufferPtr,
         *LinePtr;
   FILE  *FilePtr;
   stat_t  FstatBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadFileIntoMemory() |%s|\n", Source);

   if ((FilePtr = fopen (Source, "r", "shr=get", "shr=put")) == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      return (status);
   }

   if (fstat (fileno(FilePtr), &FstatBuffer) < 0)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fstat() %%X%08.08X\n", status);
      fclose (FilePtr);
      return (status);
   }

   Bytes = FstatBuffer.st_size;
   if (Debug) fprintf (stdout, "%d bytes\n", Bytes);
   /* a little margin for error ;^) */
   Bytes += 32;

   BufferPtr = calloc (Bytes, 1);
   if (BufferPtr == NULL)
   {
      CgiLibResponseError (FI_LI, vaxc$errno, "calloc()");
      exit (SS$_NORMAL);
   }
   BytesRemaining = Bytes;
   LinePtr = BufferPtr;

   BufferCount = 0;
   while (fgets (LinePtr, BytesRemaining, FilePtr) != NULL)
   {
      /** if (Debug) fprintf (stdout, "|%s|\n", LinePtr); **/
      Length = strlen(LinePtr);
      LinePtr += Length;
      BufferCount += Length;
      BytesRemaining -= Length;
   }
   fclose (FilePtr);

   if (Debug) fprintf (stdout, "%d |%s|\n", BufferCount, BufferPtr);

   if (FileTextPtr != NULL) *FileTextPtr = BufferPtr;
   if (FileSizePtr != NULL) *FileSizePtr = BufferCount;

   return (SS$_NORMAL);
}

/****************************************************************************/
/*
Supplied with an error number returned from a SWISH-E library call return a
pointer to an (possibly language-specific) explanation.  Error codes and
explanations derived from error codes from SWISH-E [src]error.h.
*/ 

char* SwishEsiError (int ErrorNumber)
{
   static char  UnknownString [64];

   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SwishEsiError() %d\n", ErrorNumber);

   switch (ErrorNumber)
   {
      case INDEX_FILE_NOT_FOUND :
         cptr = LANG_SwishIndexNotFound;
         break;
      case UNKNOWN_INDEX_FILE_FORMAT :
         cptr = LANG_SwishIndexFormatUnknown;
         break;
      case WORDS_TOO_COMMON :
         cptr = LANG_SwishSearchWordsCommon;
         break;
      case INDEX_FILE_IS_EMPTY :
         cptr = LANG_SwishIndexIsEmpty;
         break;
      case NO_WORDS_IN_SEARCH :
         cptr = LANG_SwishNoSearchWords;
         break;
      case UNKNOWN_PROPERTY_NAME_IN_SEARCH_DISPLAY :
         cptr = LANG_SwishSearchProperties;
         break;
      case UNKNOWN_PROPERTY_NAME_IN_SEARCH_SORT :
         cptr = LANG_SwishSearchSort;
         break;
      case UNKNOWN_METANAME :
         cptr = LANG_SwishUnknownMetaName;
         break;
      case UNIQUE_WILDCARD_NOT_ALLOWED_IN_WORD :
         cptr = LANG_SwishWildcard;
         break;
      case WORD_NOT_FOUND :
         cptr = LANG_SwishWordNotFound;
         break;
      case SWISH_LISTRESULTS_EOF :
         cptr = LANG_SwishListResultsEOF;
         break;
      case INVALID_SWISH_HANDLE :
         cptr = LANG_SwishInvalidHandle;
         break;
      default :
         sprintf (cptr = UnknownString, LANG_SwishUnknownError,
                  SWISH_VERSION, ErrorNumber);
   }

   return (cptr);
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.  OSU scripts have
the 'method', 'url' and 'protocol' supplied as P1, P2, P3 (these being detected
and used by CGILIB), and are of no interest to this function.
*/

GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   int  status,
        SkipParameters;
   unsigned short  Length;
   char  ch;
   char  *aptr, *cptr, *clptr, *sptr;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if ((clptr = getenv ("SWISHESI$PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   /* if OSU environment then skip P1, P2, P3 */
   if (getenv ("WWWEXEC_RUNDOWN_STRING") != NULL)
      SkipParameters = 3;
   else
      SkipParameters = 0;

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (SkipParameters)
      {
         SkipParameters--;
         continue;
      }

      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = true;
         continue;
      }
      if (strsame (aptr, "/INDEX=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         if (!*cptr) continue;
         CliIndexFileNamePtr = cptr;
         continue;
      }
      if (strsame (aptr, "/RESULT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         /** if (!*cptr) continue; **/
         CliResultPagePtr = cptr;
         continue;
      }
      if (strsame (aptr, "/VERSION", 4))
      {
         fprintf (stdout, "%%%s-I-SOFTWAREID, %s\n%s\n",
                  Utility, SoftwareID, CopyrightInfo);
         exit (SS$_NORMAL);
      }
      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
               Utility, aptr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strsame
(
char *sptr1,
char *sptr2,
int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/

